@@ -23,16 +23,19 @@ along with this library; if not, write to the Free Software Foundation, Inc.,
 
 #if defined(__WIN32__) || defined(_WIN32) || defined(_QNX4)
 #define _strncasecmp _strnicmp
+#define snprintf _snprintf
 #else
 #include <signal.h>
 #define USE_SIGNALS 1
 #define _strncasecmp strncasecmp
 #endif
+#include <time.h> // for "strftime()" and "gmtime()"
 
 ////////// RTSPServer //////////
 
 RTSPServer*
-RTSPServer::createNew(UsageEnvironment& env, Port ourPort) {
+RTSPServer::createNew(UsageEnvironment& env, Port ourPort,
+		      UserAuthenticationDatabase* authDatabase) {
   int ourSocket = -1;
   RTSPServer* newServer = NULL;
 
@@ -40,7 +43,7 @@ RTSPServer::createNew(UsageEnvironment& env, Port ourPort) {
     int ourSocket = setUpOurSocket(env, ourPort);
     if (ourSocket == -1) break;
 
-    return new RTSPServer(env, ourSocket, ourPort);
+    return new RTSPServer(env, ourSocket, ourPort, authDatabase);
   } while (0);
 
   if (ourSocket != -1) ::_close(ourSocket);
@@ -49,15 +52,15 @@ RTSPServer::createNew(UsageEnvironment& env, Port ourPort) {
 }
 
 Boolean RTSPServer::lookupByName(UsageEnvironment& env,
-				 char const* instanceName,
+				 char const* name,
 				 RTSPServer*& resultServer) {
   resultServer = NULL; // unless we succeed
 
   Medium* medium;
-  if (!Medium::lookupByName(env, instanceName, medium)) return False;
+  if (!Medium::lookupByName(env, name, medium)) return False;
 
   if (!medium->isRTSPServer()) {
-    env.setResultMsg(instanceName, " is not a RTSP server");
+    env.setResultMsg(name, " is not a RTSP server");
     return False;
   }
 
@@ -65,21 +68,35 @@ Boolean RTSPServer::lookupByName(UsageEnvironment& env,
   return True;
 }
 
-void RTSPServer
-::addServerMediaSession(ServerMediaSession* serverMediaSession) {
+void RTSPServer::addServerMediaSession(ServerMediaSession* serverMediaSession) {
+  if (serverMediaSession == NULL) return;
+
   char const* sessionName = serverMediaSession->streamName();
   if (sessionName == NULL) sessionName = "";
   ServerMediaSession* existingSession
     = (ServerMediaSession*)
     (fServerMediaSessions->Add(sessionName,
 			       (void*)serverMediaSession));
-  delete existingSession; // if any
+  removeServerMediaSession(existingSession); // if any
+}
+
+void RTSPServer::removeServerMediaSession(ServerMediaSession* serverMediaSession) {
+  if (serverMediaSession == NULL) return;
+
+  fServerMediaSessions->Remove(serverMediaSession->streamName());
+  if (serverMediaSession->referenceCount() == 0) {
+    delete serverMediaSession;
+  } else {
+    serverMediaSession->deleteWhenUnreferenced() = True;
+  }
 }
 
 char* RTSPServer
 ::rtspURL(ServerMediaSession const* serverMediaSession) const {
   struct in_addr ourAddress;
-  ourAddress.s_addr = ourSourceAddressForMulticast(envir()); // hack
+  ourAddress.s_addr = ReceivingInterfaceAddr != 0
+    ? ReceivingInterfaceAddr
+    : ourSourceAddressForMulticast(envir()); // hack
 
   char const* sessionName = serverMediaSession->streamName();
   unsigned sessionNameLength = strlen(sessionName);
@@ -133,9 +150,11 @@ int RTSPServer::setUpOurSocket(UsageEnvironment& env, Port& ourPort) {
 }
 
 RTSPServer::RTSPServer(UsageEnvironment& env,
-		       int ourSocket, Port ourPort)
+		       int ourSocket, Port ourPort,
+		       UserAuthenticationDatabase* authDatabase)
   : Medium(env),
     fServerSocket(ourSocket), fServerPort(ourPort),
+    fAuthDB(authDatabase),
     fServerMediaSessions(HashTable::create(STRING_HASH_KEYS)), 
     fSessionIdCounter(0) {
 #ifdef USE_SIGNALS
@@ -156,12 +175,12 @@ RTSPServer::~RTSPServer() {
 
   ::_close(fServerSocket);
 
-  // Delete all server media sessions:
+  // Remove all server media sessions (they'll get deleted when they're finished):
   while (1) {
-    ServerMediaSession* namedSession
+    ServerMediaSession* serverMediaSession
       = (ServerMediaSession*)fServerMediaSessions->RemoveNext();
-    if (namedSession == NULL) break;
-    delete namedSession;
+    if (serverMediaSession == NULL) break;
+    removeServerMediaSession(serverMediaSession);
   }
 
   // Finally, delete the session table itself:
@@ -214,8 +233,7 @@ RTSPServer::RTSPClientSession
     fSessionIsActive(True), fNumStreamStates(0), fStreamStates(NULL) {
   // Arrange to handle incoming requests:
   envir().taskScheduler().turnOnBackgroundReadHandling(fClientSocket,
-     (TaskScheduler::BackgroundHandlerProc*)&incomingRequestHandler,
-						   this);
+     (TaskScheduler::BackgroundHandlerProc*)&incomingRequestHandler, this);
 }
 
 RTSPServer::RTSPClientSession::~RTSPClientSession() {
@@ -225,12 +243,21 @@ RTSPServer::RTSPClientSession::~RTSPClientSession() {
   ::_close(fClientSocket);
 
   reclaimStreamStates();
+
+  if (fOurServerMediaSession != NULL) {
+    fOurServerMediaSession->decrementReferenceCount();
+    if (fOurServerMediaSession->referenceCount() == 0
+	&& fOurServerMediaSession->deleteWhenUnreferenced()) {
+      fOurServer.removeServerMediaSession(fOurServerMediaSession);
+    }
+  }
 }
 
 void RTSPServer::RTSPClientSession::reclaimStreamStates() {
   for (unsigned i = 0; i < fNumStreamStates; ++i) {
     if (fStreamStates[i].subsession != NULL) {
-      fStreamStates[i].subsession->deleteStream(fStreamStates[i].streamToken);
+      fStreamStates[i].subsession->deleteStream(fOurSessionId,
+						fStreamStates[i].streamToken);
     }
   }
   delete[] fStreamStates; fStreamStates = NULL;
@@ -266,6 +293,7 @@ void RTSPServer::RTSPClientSession::incomingRequestHandler1() {
       return;
     }
 #ifdef DEBUG
+    ptr[bytesRead] = '\0';
     fprintf(stderr, "RTSPClientSession[%p]::incomingRequestHandler1() read %d bytes:%s\n", this, bytesRead, ptr);
 #endif
 
@@ -311,13 +339,14 @@ void RTSPServer::RTSPClientSession::incomingRequestHandler1() {
     if (strcmp(cmdName, "OPTIONS") == 0) {
       handleCmd_OPTIONS(cseq);
     } else if (strcmp(cmdName, "DESCRIBE") == 0) {
-      handleCmd_DESCRIBE(cseq, urlSuffix);
+      handleCmd_DESCRIBE(cseq, urlSuffix, (char const*)fBuffer);
     } else if (strcmp(cmdName, "SETUP") == 0) {
       handleCmd_SETUP(cseq, urlPreSuffix, urlSuffix, (char const*)fBuffer);
     } else if (strcmp(cmdName, "TEARDOWN") == 0
 	       || strcmp(cmdName, "PLAY") == 0
 	       || strcmp(cmdName, "PAUSE") == 0) {
-      handleCmd_withinSession(cmdName, urlPreSuffix, urlSuffix, cseq);
+      handleCmd_withinSession(cmdName, urlPreSuffix, urlSuffix, cseq,
+			      (char const*)fBuffer);
     } else {
       handleCmd_notSupported(cseq);
     }
@@ -333,42 +362,60 @@ void RTSPServer::RTSPClientSession::incomingRequestHandler1() {
 
 // Handler routines for specific RTSP commands:
 
+// Generate a "Date:" header for use in a RTSP response:
+static char const* dateHeader() {
+  static char buf[200];
+  time_t tt = time(NULL);
+  strftime(buf, sizeof buf, "Date: %a, %b %d %Y %H:%M:%S GMT\r\n", gmtime(&tt));
+  return buf;
+}
+
 static char const* allowedCommandNames
   = "OPTIONS, DESCRIBE, SETUP, TEARDOWN, PLAY, PAUSE";
 
 void RTSPServer::RTSPClientSession::handleCmd_bad(char const* /*cseq*/) {
   // Don't do anything with "cseq", because it might be nonsense
-  sprintf((char*)fBuffer, "RTSP/1.0 400 Bad Request\r\nAllow: %s\r\n\r\n",
-	  allowedCommandNames);
+  snprintf((char*)fBuffer, sizeof fBuffer,
+	   "RTSP/1.0 400 Bad Request\r\n%sAllow: %s\r\n\r\n",
+	   dateHeader(), allowedCommandNames);
   fSessionIsActive = False; // triggers deletion of ourself after responding
 }
 
 void RTSPServer::RTSPClientSession::handleCmd_notSupported(char const* cseq) {
-  sprintf((char*)fBuffer, "RTSP/1.0 405 Method Not Allowed\r\nCSeq: %s\r\nAllow: %s\r\n\r\n",
-	  cseq, allowedCommandNames);
+  snprintf((char*)fBuffer, sizeof fBuffer,
+	   "RTSP/1.0 405 Method Not Allowed\r\nCSeq: %s\r\n%sAllow: %s\r\n\r\n",
+	   cseq, dateHeader(), allowedCommandNames);
   fSessionIsActive = False; // triggers deletion of ourself after responding
 }
 
 void RTSPServer::RTSPClientSession::handleCmd_notFound(char const* cseq) {
-  sprintf((char*)fBuffer, "RTSP/1.0 404 Stream Not Found\r\nCSeq: %s\r\n\r\n", cseq);
+  snprintf((char*)fBuffer, sizeof fBuffer,
+	   "RTSP/1.0 404 Stream Not Found\r\nCSeq: %s\r\n%s\r\n",
+	   cseq, dateHeader());
   fSessionIsActive = False; // triggers deletion of ourself after responding
 }
 
 void RTSPServer::RTSPClientSession::handleCmd_unsupportedTransport(char const* cseq) {
-  sprintf((char*)fBuffer, "RTSP/1.0 461 Unsupported Transport\r\nCSeq: %s\r\n\r\n", cseq);
+  snprintf((char*)fBuffer, sizeof fBuffer,
+	   "RTSP/1.0 461 Unsupported Transport\r\nCSeq: %s\r\n%s\r\n",
+	   cseq, dateHeader());
   fSessionIsActive = False; // triggers deletion of ourself after responding
 }
 
 void RTSPServer::RTSPClientSession::handleCmd_OPTIONS(char const* cseq) {
-  sprintf((char*)fBuffer, "RTSP/1.0 200 OK\r\nCSeq: %s\r\nPublic: %s\r\n\r\n",
-	  cseq, allowedCommandNames);
+  snprintf((char*)fBuffer, sizeof fBuffer,
+	   "RTSP/1.0 200 OK\r\nCSeq: %s\r\n%sPublic: %s\r\n\r\n",
+	   cseq, dateHeader(), allowedCommandNames);
 }
 
 void RTSPServer::RTSPClientSession
-::handleCmd_DESCRIBE(char const* cseq, char const* urlSuffix) {
+::handleCmd_DESCRIBE(char const* cseq, char const* urlSuffix,
+		     char const* fullRequestStr) {
   char* sdpDescription = NULL;
   char* rtspURL = NULL;
   do {
+    if (!authenticationOK("DESCRIBE", cseq, fullRequestStr)) break;
+
     // We should really check that the request contains an "Accept:" #####
     // for "application/sdp", because that's what we're sending back #####
 
@@ -386,7 +433,12 @@ void RTSPServer::RTSPClientSession
     if (sdpDescription == NULL) {
       // This usually means that a file name that was specified for a
       // "ServerMediaSubsession" does not exist.
-      sprintf((char*)fBuffer, "RTSP/1.0 404 File Not Found, Or In Incorrect Format\r\nCSeq: %s\r\n\r\n", cseq);
+      snprintf((char*)fBuffer, sizeof fBuffer,
+	       "RTSP/1.0 404 File Not Found, Or In Incorrect Format\r\n"
+	       "CSeq: %s\r\n"
+	       "%s\r\n",
+	       cseq,
+	       dateHeader());
      break;
     }
     unsigned sdpDescriptionSize = strlen(sdpDescription);
@@ -395,24 +447,19 @@ void RTSPServer::RTSPClientSession
     // (which is necessary to ensure that the correct URL gets used in
     // subsequent "SETUP" requests).
     rtspURL = fOurServer.rtspURL(session);
-    unsigned rtspURLSize = strlen(rtspURL); 
 
-    if (sdpDescriptionSize + rtspURLSize > sizeof fBuffer - 200) { // sanity check
-      sprintf((char*)fBuffer,
-	      "RTSP/1.0 500 Internal Server Error\r\nCSeq: %s\r\n\r\n", cseq);
-      break;
-    }
-  
-    sprintf((char*)fBuffer,
-	    "RTSP/1.0 200 OK\r\nCSeq: %s\r\n"
-	    "Content-Base: %s/\r\n"
-	    "Content-Type: application/sdp\r\n"
-	    "Content-Length: %d\r\n\r\n"
-	    "%s",
-	    cseq,
-	    rtspURL,
-	    sdpDescriptionSize,
-	    sdpDescription);
+    snprintf((char*)fBuffer, sizeof fBuffer,
+	     "RTSP/1.0 200 OK\r\nCSeq: %s\r\n"
+	     "%s"
+	     "Content-Base: %s/\r\n"
+	     "Content-Type: application/sdp\r\n"
+	     "Content-Length: %d\r\n\r\n"
+	     "%s",
+	     cseq,
+	     dateHeader(),
+	     rtspURL,
+	     sdpDescriptionSize,
+	     sdpDescription);
   } while (0);
 
   delete[] sdpDescription;
@@ -428,7 +475,7 @@ static void parseTransportHeader(char const* buf,
 				 unsigned char& rtpChannelId, // if TCP
 				 unsigned char& rtcpChannelId // if TCP
 				 ) {
-  // Initialize the return parameters to default values:
+  // Initialize the result parameters to default values:
   tcpStreamingRequested = False;
   destinationAddressStr = NULL;
   destinationTTL = 255;
@@ -466,8 +513,8 @@ static void parseTransportHeader(char const* buf,
     }
 
     fields += strlen(field);
-    while (fields[0] == ';') ++fields; // skip over all leading ';' chars
-    if (fields[0] == '\0') break;
+    while (*fields == ';') ++fields; // skip over separating ';' chars
+    if (*fields == '\0' || *fields == '\r' || *fields == '\n') break;
   }
   delete[] field;
 }
@@ -497,6 +544,8 @@ void RTSPServer::RTSPClientSession
       return;
     }
 
+    fOurServerMediaSession->incrementReferenceCount();
+
     // Set up our array of states for this session's subsessions (tracks):
     reclaimStreamStates();
     ServerMediaSubsessionIterator iter(*fOurServerMediaSession);
@@ -571,33 +620,39 @@ void RTSPServer::RTSPClientSession
       handleCmd_unsupportedTransport(cseq);
       return;
     }
-    sprintf((char*)fBuffer,
-	    "RTSP/1.0 200 OK\r\n"
-	    "CSeq: %s\r\n"
-	    "Transport: RTP/AVP;multicast;destination=%s;port=%d;ttl=%d\r\n"
-	    "Session: %d\r\n\r\n",
-	    cseq,
-	    our_inet_ntoa(destinationAddr), ntohs(serverRTPPort.num()), destinationTTL,
-	    fOurSessionId);
+    snprintf((char*)fBuffer, sizeof fBuffer,
+	     "RTSP/1.0 200 OK\r\n"
+	     "CSeq: %s\r\n"
+	     "%s"
+	     "Transport: RTP/AVP;multicast;destination=%s;port=%d;ttl=%d\r\n"
+	     "Session: %d\r\n\r\n",
+	     cseq,
+	     dateHeader(),
+	     our_inet_ntoa(destinationAddr), ntohs(serverRTPPort.num()), destinationTTL,
+	     fOurSessionId);
   } else {
     if (tcpStreamingRequested) {
-      sprintf((char*)fBuffer,
-	      "RTSP/1.0 200 OK\r\n"
-	      "CSeq: %s\r\n"
-	      "Transport: RTP/AVP/TCP;unicast;destination=%s;interleaved=%d-%d\r\n"
-	      "Session: %d\r\n\r\n",
-	      cseq,
-	      our_inet_ntoa(destinationAddr), rtpChannelId, rtcpChannelId,
-	      fOurSessionId);
+      snprintf((char*)fBuffer, sizeof fBuffer,
+	       "RTSP/1.0 200 OK\r\n"
+	       "CSeq: %s\r\n"
+	       "%s"
+	       "Transport: RTP/AVP/TCP;unicast;destination=%s;interleaved=%d-%d\r\n"
+	       "Session: %d\r\n\r\n",
+	       cseq,
+	       dateHeader(),
+	       our_inet_ntoa(destinationAddr), rtpChannelId, rtcpChannelId,
+	       fOurSessionId);
     } else {
-      sprintf((char*)fBuffer,
-	      "RTSP/1.0 200 OK\r\n"
-	      "CSeq: %s\r\n"
-	      "Transport: RTP/AVP;unicast;destination=%s;client_port=%d-%d;server_port=%d-%d\r\n"
-	      "Session: %d\r\n\r\n",
-	      cseq,
-	      our_inet_ntoa(destinationAddr), ntohs(clientRTPPort.num()), ntohs(clientRTCPPort.num()), ntohs(serverRTPPort.num()), ntohs(serverRTCPPort.num()),
-	      fOurSessionId);
+      snprintf((char*)fBuffer, sizeof fBuffer,
+	       "RTSP/1.0 200 OK\r\n"
+	       "CSeq: %s\r\n"
+	       "%s"
+	       "Transport: RTP/AVP;unicast;destination=%s;client_port=%d-%d;server_port=%d-%d\r\n"
+	       "Session: %d\r\n\r\n",
+	       cseq,
+	       dateHeader(),
+	       our_inet_ntoa(destinationAddr), ntohs(clientRTPPort.num()), ntohs(clientRTCPPort.num()), ntohs(serverRTPPort.num()), ntohs(serverRTCPPort.num()),
+	       fOurSessionId);
     }
   }
 }
@@ -605,7 +660,7 @@ void RTSPServer::RTSPClientSession
 void RTSPServer::RTSPClientSession
 ::handleCmd_withinSession(char const* cmdName,
 			  char const* urlPreSuffix, char const* urlSuffix,
-			  char const* cseq) {
+			  char const* cseq, char const* fullRequestStr) {
   // This will either be:
   // - a non-aggregated operation, if "urlPreSuffix" is the session (stream)
   //   name and "urlSuffix" is the subsession (track) name, or
@@ -642,34 +697,159 @@ void RTSPServer::RTSPClientSession
   if (strcmp(cmdName, "TEARDOWN") == 0) {
     handleCmd_TEARDOWN(subsession, cseq);
   } else if (strcmp(cmdName, "PLAY") == 0) {
-    handleCmd_PLAY(subsession, cseq);
+    handleCmd_PLAY(subsession, cseq, fullRequestStr);
   } else if (strcmp(cmdName, "PAUSE") == 0) {
     handleCmd_PAUSE(subsession, cseq);
   }
 }
 
 void RTSPServer::RTSPClientSession
-  ::handleCmd_TEARDOWN(ServerMediaSubsession* subsession, char const* cseq) {
-  for (unsigned i = 0; i < fNumStreamStates; ++i) {
-    if (subsession == NULL /* means: aggregated operation */
-	|| subsession == fStreamStates[i].subsession) {
-      fStreamStates[i].subsession->endStream(fStreamStates[i].streamToken);
-    }
-  }
-  sprintf((char*)fBuffer, "RTSP/1.0 200 OK\r\nCSeq: %s\r\n\r\n", cseq);
+::handleCmd_TEARDOWN(ServerMediaSubsession* /*subsession*/, char const* cseq) {
+  snprintf((char*)fBuffer, sizeof fBuffer,
+	   "RTSP/1.0 200 OK\r\nCSeq: %s\r\n%s\r\n",
+	   cseq, dateHeader());
   fSessionIsActive = False; // triggers deletion of ourself after responding
 }
 
+static Boolean parseRangeHeader(char const* buf, float& rangeStart, float& rangeEnd) {
+  // Initialize the result parameters to default values:
+  rangeStart = rangeEnd = 0.0;
+
+  // First, find "Range:"
+  while (1) {
+    if (*buf == '\0') return False; // not found
+    if (_strncasecmp(buf, "Range: ", 7) == 0) break;
+    ++buf;
+  }
+
+  // Then, run through each of the fields, looking for ones we handle:
+  char const* fields = buf + 7;
+  while (*fields == ' ') ++fields;
+  float start, end;
+  if (sscanf(fields, "npt = %f - %f", &start, &end) == 2) {
+    rangeStart = start;
+    rangeEnd = end;
+  } else if (sscanf(fields, "npt = %f -", &start) == 1) {
+    rangeStart = start;
+  } else {
+    return False; // The header is malformed
+  }
+
+  return True;
+}
+
 void RTSPServer::RTSPClientSession
-  ::handleCmd_PLAY(ServerMediaSubsession* subsession, char const* cseq) {
+  ::handleCmd_PLAY(ServerMediaSubsession* subsession, char const* cseq,
+		   char const* fullRequestStr) {
+  char* rtspURL = fOurServer.rtspURL(fOurServerMediaSession);
+  unsigned rtspURLSize = strlen(rtspURL);
+
+  // Parse the client's "Range:" header, if any: 
+  float rangeStart, rangeEnd;
+  Boolean sawRangeHeader = parseRangeHeader(fullRequestStr, rangeStart, rangeEnd);
+
+  // Use this information, plus the stream's duration (if known), to create
+  // our own "Range:" header, for the response:
+  float duration = subsession == NULL /*aggregate op*/
+    ? fOurServerMediaSession->duration() : subsession->duration();
+  if (duration < 0.0) {
+    // We're an aggregate PLAY, but the subsessions have different durations.
+    // Use the largest of these durations in our header
+    duration = -duration;
+  }
+
+  if (rangeEnd < 0.0 || rangeEnd > duration) rangeEnd = duration;
+  if (rangeStart < 0.0) {
+    rangeStart = 0.0;
+  } else if (rangeEnd > 0.0 && rangeStart > rangeEnd) {
+    rangeStart = rangeEnd;
+  }
+
+  char buf[100];
+  char* rangeHeader;
+  if (!sawRangeHeader) {
+    buf[0] = '\0'; // Because we didn't see a Range: header, don't send one back
+  } else if (rangeEnd == 0.0) {
+    sprintf(buf, "Range: npt=%.3f-\r\n", rangeStart);
+  } else {
+    sprintf(buf, "Range: npt=%.3f-%.3f\r\n", rangeStart, rangeEnd);
+  }
+  rangeHeader = strDup(buf);
+
+  // Create a "RTP-Info:" line:
+  char const* rtpInfoFmt =
+    "%s" // "RTP-Info:", plus any preceding rtpInfo items
+    "%s" // comma separator, if needed 
+    "url=%s/%s"
+    ";seq=%d"
+#ifdef RTPINFO_INCLUDE_RTPTIME
+    ";rtptime=%d"
+#endif
+    ;
+  unsigned rtpInfoFmtSize = strlen(rtpInfoFmt);
+  char* rtpInfo = strDup("RTP-Info: ");
+  unsigned numRTPInfoItems = 0;
+
   for (unsigned i = 0; i < fNumStreamStates; ++i) {
     if (subsession == NULL /* means: aggregated operation */
 	|| subsession == fStreamStates[i].subsession) {
-      fStreamStates[i].subsession->startStream(fStreamStates[i].streamToken);
+      unsigned short rtpSeqNum = 0;
+      unsigned rtpTimestamp = 0;
+      if (sawRangeHeader) {
+	fStreamStates[i].subsession->seekStream(fOurSessionId,
+						fStreamStates[i].streamToken,
+						rangeStart);
+      }
+      fStreamStates[i].subsession->startStream(fOurSessionId,
+					       fStreamStates[i].streamToken,
+					       rtpSeqNum, rtpTimestamp);
+      const char *urlSuffix = fStreamStates[i].subsession->trackId();
+      char* prevRTPInfo = rtpInfo;
+      unsigned rtpInfoSize = rtpInfoFmtSize
+	+ strlen(prevRTPInfo)
+	+ 1
+	+ rtspURLSize + strlen(urlSuffix)
+	+ 5 /*max unsigned short len*/
+#ifdef RTPINFO_INCLUDE_RTPTIME
+	+ 10 /*max unsigned (32-bit) len*/
+#endif
+	+ 2 /*allows for trailing \r\n at final end of string*/; 
+      rtpInfo = new char[rtpInfoSize];
+      sprintf(rtpInfo, rtpInfoFmt,
+	      prevRTPInfo,
+	      numRTPInfoItems++ == 0 ? "" : ",",
+	      rtspURL, urlSuffix,
+	      rtpSeqNum
+#ifdef RTPINFO_INCLUDE_RTPTIME
+	      ,rtpTimestamp
+#endif
+	      );
+      delete[] prevRTPInfo;
     }
   }
-  sprintf((char*)fBuffer, "RTSP/1.0 200 OK\r\nCSeq: %s\r\nSession: %d\r\n\r\n",
-	  cseq, fOurSessionId);
+  if (numRTPInfoItems == 0) {
+    rtpInfo[0] = '\0';
+  } else {
+    unsigned rtpInfoLen = strlen(rtpInfo);
+    rtpInfo[rtpInfoLen] = '\r';
+    rtpInfo[rtpInfoLen+1] = '\n';
+    rtpInfo[rtpInfoLen+2] = '\0';
+  }
+
+  // Fill in the response:
+  snprintf((char*)fBuffer, sizeof fBuffer,
+	   "RTSP/1.0 200 OK\r\n"
+	   "CSeq: %s\r\n"
+	   "%s"
+	   "%s"
+	   "Session: %d\r\n"
+	   "%s\r\n",
+	   cseq,
+	   dateHeader(),
+	   rangeHeader,
+	   fOurSessionId,
+	   rtpInfo);
+  delete[] rtpInfo; delete[] rangeHeader;
 }
 
 void RTSPServer::RTSPClientSession
@@ -677,11 +857,123 @@ void RTSPServer::RTSPClientSession
   for (unsigned i = 0; i < fNumStreamStates; ++i) {
     if (subsession == NULL /* means: aggregated operation */
 	|| subsession == fStreamStates[i].subsession) {
-      fStreamStates[i].subsession->pauseStream(fStreamStates[i].streamToken);
+      fStreamStates[i].subsession->pauseStream(fOurSessionId,
+					       fStreamStates[i].streamToken);
     }
   }
-  sprintf((char*)fBuffer, "RTSP/1.0 200 OK\r\nCSeq: %s\r\nSession: %d\r\n\r\n",
-	  cseq, fOurSessionId);
+  snprintf((char*)fBuffer, sizeof fBuffer,
+	   "RTSP/1.0 200 OK\r\nCSeq: %s\r\n%sSession: %d\r\n\r\n",
+	   cseq, dateHeader(), fOurSessionId);
+}
+
+static Boolean parseAuthorizationHeader(char const* buf,
+					char const*& username,
+					char const*& realm,
+					char const*& nonce, char const*& uri,
+					char const*& response) {
+  // Initialize the result parameters to default values:
+  username = realm = nonce = uri = response = NULL;
+
+  // First, find "Authorization:"
+  while (1) {
+    if (*buf == '\0') return False; // not found
+    if (_strncasecmp(buf, "Authorization: Digest ", 22) == 0) break;
+    ++buf;
+  }
+
+  // Then, run through each of the fields, looking for ones we handle:
+  char const* fields = buf + 22;
+  while (*fields == ' ') ++fields;
+  char* parameter = strDupSize(fields);
+  char* value = strDupSize(fields);
+  while (1) {
+    value[0] = '\0';
+    if (sscanf(fields, "%[^=]=\"%[^\"]\"", parameter, value) != 2 &&
+	sscanf(fields, "%[^=]=\"\"", parameter) != 1) {
+      break;
+    }
+    if (strcmp(parameter, "username") == 0) {
+      username = strDup(value);
+    } else if (strcmp(parameter, "realm") == 0) {
+      realm = strDup(value);
+    } else if (strcmp(parameter, "nonce") == 0) {
+      nonce = strDup(value);
+    } else if (strcmp(parameter, "uri") == 0) {
+      uri = strDup(value);
+    } else if (strcmp(parameter, "response") == 0) {
+      response = strDup(value);
+    }
+
+    fields += strlen(parameter) + 2 /*="*/ + strlen(value) + 1 /*"*/;
+    while (*fields == ',' || *fields == ' ') ++fields;
+        // skip over any separating ',' and ' ' chars
+    if (*fields == '\0' || *fields == '\r' || *fields == '\n') break;
+  }
+  delete[] parameter; delete[] value;
+  return True;
+}
+
+Boolean RTSPServer::RTSPClientSession
+::authenticationOK(char const* cmdName, char const* cseq,
+		   char const* fullRequestStr) {
+  // If we weren't set up with an authentication database, we're OK:
+  if (fOurServer.fAuthDB == NULL) return True;
+
+  char const* username = NULL; char const* realm = NULL; char const* nonce = NULL;
+  char const* uri = NULL; char const* response = NULL;
+  Boolean success = False;
+
+  do {
+    // To authenticate, we first need to have a nonce set up
+    // from a previous attempt:
+    if (fCurrentAuthenticator.nonce() == NULL) break;
+
+    // Next, the request needs to contain an "Authorization:" header,
+    // containing a username, (our) realm, (our) nonce, uri,
+    // and response string:
+    if (!parseAuthorizationHeader(fullRequestStr,
+				  username, realm, nonce, uri, response)
+	|| username == NULL
+	|| realm == NULL || strcmp(realm, fCurrentAuthenticator.realm()) != 0
+	|| nonce == NULL || strcmp(nonce, fCurrentAuthenticator.nonce()) != 0
+	|| uri == NULL || response == NULL) {
+      break;
+    }
+
+    // Next, the username has to be known to us:
+    char const* password = fOurServer.fAuthDB->lookupPassword(username);
+#ifdef DEBUG
+    fprintf(stderr, "lookupPassword(%s) returned password %s\n", username, password);
+#endif
+    if (password == NULL) break;
+    fCurrentAuthenticator.
+      setUsernameAndPassword(username, password,
+			     fOurServer.fAuthDB->passwordsAreMD5());
+
+    // Finally, compute a digest response from the information that we have,
+    // and compare it to the one that we were given:
+    char const* ourResponse
+      = fCurrentAuthenticator.computeDigestResponse(cmdName, uri);
+    success = (strcmp(ourResponse, response) == 0);
+    fCurrentAuthenticator.reclaimDigestResponse(ourResponse);
+  } while (0);
+
+  delete[] (char*)username; delete[] (char*)realm; delete[] (char*)nonce;
+  delete[] (char*)uri; delete[] (char*)response;
+  if (success) return True;
+
+  // If we get here, there was some kind of authentication failure.
+  // Send back a "401 Unauthorized" response, with a new random nonce:
+  fCurrentAuthenticator.setRealmAndRandomNonce(fOurServer.fAuthDB->realm());
+  snprintf((char*)fBuffer, sizeof fBuffer,
+	   "RTSP/1.0 401 Unauthorized\r\n"
+	   "CSeq: %s\r\n"
+	   "%s"
+	   "WWW-Authenticate: Digest realm=\"%s\", nonce=\"%s\"\r\n\r\n",
+	   cseq,
+	   dateHeader(),
+	   fCurrentAuthenticator.realm(), fCurrentAuthenticator.nonce());
+  return False;
 }
 
 Boolean
@@ -703,7 +995,7 @@ RTSPServer::RTSPClientSession
   unsigned i;
   for (i = 0; i < resultCmdNameMaxSize-1 && i < reqStrSize; ++i) {
     char c = reqStr[i];
-    if (c == ' ') {
+    if (c == ' ' || c == '\t') {
       parseSucceeded = True;
       break;
     }
@@ -715,7 +1007,7 @@ RTSPServer::RTSPClientSession
       
   // Skip over the prefix of any "rtsp://" URL that follows:
   unsigned j = i+1;
-  while (j < reqStrSize && reqStr[j] == ' ') ++j; // skip over any additional spaces
+  while (j < reqStrSize && (reqStr[j] == ' ' || reqStr[j] == '\t')) ++j; // skip over any additional white space
   for (j = i+1; j < reqStrSize-8; ++j) {
     if ((reqStr[j] == 'r' || reqStr[j] == 'R')
 	&& (reqStr[j+1] == 't' || reqStr[j+1] == 'T')
@@ -787,3 +1079,33 @@ RTSPServer::RTSPClientSession
 
   return True;
 }
+
+
+////////// UserAuthenticationDatabase implementation //////////
+
+UserAuthenticationDatabase::UserAuthenticationDatabase(char const* realm,
+						       Boolean passwordsAreMD5)
+  : fTable(HashTable::create(STRING_HASH_KEYS)),
+    fRealm(strDup(realm == NULL ? "LIVE.COM Streaming Media" : realm)),
+    fPasswordsAreMD5(passwordsAreMD5) {
+}
+
+UserAuthenticationDatabase::~UserAuthenticationDatabase() {
+  delete[] fRealm;
+  delete fTable;
+}
+
+void UserAuthenticationDatabase::addUserRecord(char const* username,
+					       char const* password) {
+  fTable->Add(username, (void*)(strDup(password)));
+}
+
+void UserAuthenticationDatabase::removeUserRecord(char const* username) {
+  char* password = (char*)(fTable->Lookup(username));
+  fTable->Remove(username);
+  delete[] password;
+}
+
+char const* UserAuthenticationDatabase::lookupPassword(char const* username) {
+  return (char const*)(fTable->Lookup(username));
+}
